if not info.fatality.allow_insecure then
    return error("The images library requires unsafe scripts to be enabled!")
end


local function vtable_bind(class, _type, index)
    local this      = ffi.cast("void***", class)
    local ffitype   = ffi.typeof(_type)
    return function (...)
        if this ~= 0 and this ~= nil then
            return ffi.cast(ffitype, this[0][index])(this, ...)
        else
            return error(string.format("vtable_bind! type: %s. index: %i", _type, index))
        end
    end
end

local bytearray = ffi.typeof("uint8_t[?]")

local function vfunc_args(return_type, args)
    return string.format("%s(__thiscall*)(void*%s)", return_type or "void", args and (", " .. args) or "")
end

ffi.cdef[[
    struct SteamApiContext_t 
    {
        void* m_pSteamClient;
        void* m_pSteamUser;
        void* m_pSteamFriends;
        void* m_pSteamUtils;
    };

    struct PlayerInfo_t
    {
        int64_t			ver;
        int64_t			xuid;
        char			name[128];
        int				userid;
        char			guid[33];
        unsigned int	friend_id;
        char			friend_name[128];
        bool			bot;
        bool			hltv;
        unsigned int	custom_files[4];
        unsigned char	files_downloaded;
    };
]]

local Engine                = utils.find_interface("engine.dll", "VEngineClient014")
local GetSteamApiContext    = vtable_bind(Engine, vfunc_args("struct SteamApiContext_t*", nil), 185)
local GetPlayerInfo         = vtable_bind(Engine, vfunc_args("bool", "int, struct PlayerInfo_t*"), 8)

local SteamContext          = GetSteamApiContext() or error("Could not find steam API context!")
local GetSmallFriendAvatar  = vtable_bind(SteamContext.m_pSteamFriends, vfunc_args("int", "uint64_t"), 34)
local GetMediumFriendAvatar = vtable_bind(SteamContext.m_pSteamFriends, vfunc_args("int", "uint64_t"), 35)
local GetLargeFriendAvatar  = vtable_bind(SteamContext.m_pSteamFriends, vfunc_args("int", "uint64_t"), 36)
local GetImageSize          = vtable_bind(SteamContext.m_pSteamUtils,   vfunc_args("bool", "int, uint32_t*, uint32_t*"), 5)
local GetImageRGBA          = vtable_bind(SteamContext.m_pSteamUtils,   vfunc_args("bool", "int, uint8_t*, int"), 6)

-- Credit - Sapphyrus
local FileSystem        = utils.find_interface("filesystem_stdio.dll", "VBaseFileSystem011")
local NativeReadFile    = vtable_bind(FileSystem, "int(__thiscall*)(void*, void*, int, void*)", 0)
local OpenFile          = vtable_bind(FileSystem, "void*(__thiscall*)(void*, const char*, const char*, const char*)", 2)
local GetFileSize       = vtable_bind(FileSystem, "unsigned int(__thiscall*)(void*, void*)", 7)

local function ReadFile(filename)
	local FileHandle = OpenFile(filename, "r", "MOD")
	if not FileHandle then
        return
    end

	local Size = GetFileSize(FileHandle)
	if not Size or Size <= 0 then
        return
    end

	local Buffer = bytearray(Size + 1)
	if not Buffer then
        return
    end

	if not NativeReadFile(Buffer, Size, FileHandle) then
        return
    end

	return Buffer, Size
end

local lib = {}

local avatar_textures = {}
local image_w, image_h = ffi.new("uint32_t[1]"), ffi.new("uint32_t[1]")
function lib.get_steam_pfp_from_id(id, size)
    if not size then
        size = 64
    end
    
    local type = 0
    if size > 64 then
        type = 2
    elseif size > 32 then
        type = 1
    end

    local str_id = tostring(id)
    if not avatar_textures[str_id] then
        avatar_textures[str_id] = {}
    end

    if not avatar_textures[str_id][type] then
        
        local AvatarID = nil
        if type == 0 then
            AvatarID = GetSmallFriendAvatar(id)
        elseif type == 1 then
            AvatarID = GetMediumFriendAvatar(id)
        else
            AvatarID = GetLargeFriendAvatar(id)
        end


        if GetImageSize(AvatarID, image_w, image_h) then
            local buffer_size = image_w[0] * image_h[0] * 4
            local buffer = bytearray(buffer_size)

            if GetImageRGBA(AvatarID, buffer, buffer_size) then
                avatar_textures[str_id][type] = render.create_texture_rgba(tonumber(ffi.cast("uint32_t", buffer)), image_w[0], image_h[0], image_w[0] * 4)
                return avatar_textures[str_id][type]
            end
        end
    else
        return avatar_textures[str_id][type]
    end
end

local pinfo = ffi.new("struct PlayerInfo_t[1]")
function lib.get_steam_pfp(player_or_id, size)

    if not player_or_id then
        return
    end

    if type(player_or_id) == "cdata" then
        return lib.get_steam_pfp_from_id(player_or_id, size)
    end

    -- check if valid, is the correct type and is a player
    if not player_or_id.is_player or not player_or_id:is_player() then
        return
    end

    if GetPlayerInfo(player_or_id:get_index(), pinfo) then
        if pinfo[0].bot then
            local teamnum = player_or_id:get_prop("m_iTeamNum")
            -- these only have one size sadly
            if teamnum == 2 then
                return lib.get_image_from_file("materials/panorama/images/icons/scoreboard/avatar-terrorist.png")
            elseif teamnum == 3 then
                return lib.get_image_from_file("materials/panorama/images/icons/scoreboard/avatar-ct.png")
            end
        else
            return lib.get_steam_pfp_from_id(pinfo[0].xuid, size)
        end
    end
end

local file_textures = {}
function lib.get_image_from_file(dir, height)
    if not file_textures[dir] then
        -- thanks sapphyrus for this pattern magic
		local bytes, size = ReadFile(string.gsub(string.gsub(string.gsub(string.gsub(string.gsub(dir, "%z", ""), "%c", ""), "\\", "/"), "%.%./", ""), "^/+", ""))

		if bytes then
            if string.find(dir, ".svg") then
                file_textures[dir] = render.create_texture_svg(ffi.string(bytes), height or 100)
            else
                local texture = render.create_texture_bytes(tonumber(ffi.cast("uint32_t", bytes)), size - 1)
                if texture ~= 0 then
                    file_textures[dir] = texture
                else
                    file_textures[dir] = nil
                end
            end
		else
			file_textures[dir] = nil
		end
	end

    return file_textures[dir]
end

function lib.get_weapon_icon(weapon, height)

    local type = type(weapon)
    local weapon_name = nil
    -- check if its a weapon opject
    if type == "userdata" and weapon.get_prop then
        local item_definition_index = weapon:get_prop("m_iItemDefinitionIndex")

        -- can be nil if wrong entity type
        if item_definition_index then
            local weapon_info = utils.get_weapon_info(item_definition_index)
            if weapon_info then
                weapon_name = string.gsub(weapon_info.console_name, "weapon_", "") .. ".svg"
            end
        end
    elseif type == "string" then
        local name      = string.gsub(weapon, " ", "")
        local str_len   = name:len()
        if name:sub(str_len - 3, str_len) ~= ".svg" then
            name = name .. ".svg"
        end

        weapon_name = name
    end

    if weapon_name then
        return lib.get_image_from_file("materials/panorama/images/icons/equipment/" .. weapon_name, height)
    end
end

return lib